import os
import sys
import math  

# Task: convert dot-bracket notation with arbitrary pseudoknots to CT file

def db_to_ct_brackets(structure):

    stack_round = []
    stack_square = []
    stack_curly = []

    base_pairs = []

    """ Scan the structure and put different types of opening brackets on different stacks,
        i.e. in three different lists. """
    for index, element in enumerate(structure):

        if element == '(':
            stack_round.append(index + 1)
        if element == ')':
            bp = (stack_round.pop(), (index + 1))
            bp_rev = (bp[1], bp[0])
            base_pairs.append(bp)
            base_pairs.append(bp_rev)
            
        if element == '[':
            stack_square.append(index + 1)
        if element == ']':
            bp = (stack_square.pop(), (index + 1))
            bp_rev = (bp[1], bp[0])
            base_pairs.append(bp)
            base_pairs.append(bp_rev)
            
        if element == '{':
            stack_curly.append(index + 1)
        if element == '}':
            bp = (stack_curly.pop(), (index + 1))
            bp_rev = (bp[1], bp[0])
            base_pairs.append(bp)
            base_pairs.append(bp_rev)
            
    return base_pairs

def write_CT_file(sequence, base_pairs, output_ct):

    output = file(output_ct,'w')

    output.write(str(len(sequence)))
    output.write('\n')    
        
    for index, base in enumerate(sequence):

        found = False
        
        for (b1, b2) in base_pairs:
            if b1 == index + 1:
                line = (index + 1, base, index, index + 2, b2, index + 1)
                found = True
                
        if not found:
            line = (index + 1, base, index, index + 2, 0, index + 1)

        for element in line:
            output.write(str(element))
            output.write(' ')
        output.write('\n')
        

    output.close()

##################### Main Program #####################
#sequence = 'UUAUACCCGCUUAAUUCAUUCAGAUCUGUAAUAGAACUGUCAUUCAACCCCAAAAAUCUAGUGCUGAUAUAACCUUCACCAAUUAGGUUCAAAUAAGUGGUAAUGCGGGACAAAAGACUAUCGACAUUUGAUACACUAUUUAUCAAUGGAUGUCUUAUUUUUUUU'
#structure = '...(((((..........(((.(((.....))).)))...............................((((...[[[[.[[[))))....]]].]]]]...)))))....(((((.(((..((((.((((.......))))))))))))))))...........'
#output_ct = 'Z:/Documents/PhD/Multiple_Alignment_DotKnot_zscores_latest/RF00025_Telomerase_Cil/RF00025.ct'
#base_pairs = db_to_ct_brackets(structure)
#write_CT_file(sequence, base_pairs, output_ct)
