"""
    DotKnot-PW: software for pairwise RNA pseudoknot prediction
    Copyright (C) 2011-2012 Jana Sperschneider	

    This program is free software; you can redistribute it and/or modify  
    it under the terms of the GNU General Public License as published by  
    the Free Software Foundation; either version 2 of the License, or     
    (at your option) any later version.    
"""
import os
import sys
import getopt
from subprocess import *
from operator import itemgetter

import main_function
import user_input
import BPs_to_CT
#---------------------------------------------------------------------------------------------------------
RNAFOLD_PATH = "./ViennaRNA-1.8.5/Progs/"
#---------------------------------------------------------------------------------------------------------
def usage():
    """ Function: usage()

        Purpose:  Print helpful information for the user.               
    """
    print 
    print "Usage for DotKnot-PW: ", 
    print "python dotknot.py <input_file>"
    print
    print "For example, run: "
    print "python dotknot.py ./TestSet_Rfam/RF00094_HDV/RF00094_10seqs.fasta"
    print    
    sys.exit(0)    

    return
#---------------------------------------------------------------------------------------------------------
commandline = sys.argv

if commandline[-1] == '-h' or commandline[-1] == '-help' or commandline[-1] == '--h' or commandline[-1] == '--help':
    usage()
    sys.exit()
#---------------------------------------------------------------------------------------------------------
list_of_ids, list_of_dictionaries = user_input.read_input()
#---------------------------------------------------------------------------------------------------------
print '-------------------'
print
print "DotKnot-PW is running..."
print
print '-------------------'
#---------------------------------------------------------------------------------------------------------
FILE_ID = list_of_ids[0].strip()
FILE_ID = FILE_ID.replace('>','')
FILE_ID = FILE_ID.split()
if len(FILE_ID) > 1:
    FILE_ID = FILE_ID[0]
else:
    FILE_ID = str(FILE_ID[0])
FILE_ID = FILE_ID.replace('.','')
FILE_ID = FILE_ID.replace('_','')
FILE_ID = FILE_ID.replace('/','')
FILE_ID = FILE_ID.replace('|','')
FILE_ID = FILE_ID[:12]
#---------------------------------------------------------------------------------------------------------
path = commandline[-1].replace(".fasta", "" )
output_alignment = file(path + "_output.txt",'w')
#---------------------------------------------------------------------------------------------------------
# Compute dictionaries for first sequences
i = 0
seq1 = list_of_dictionaries[i]
stem_dic1 = main_function.find_stems(seq1, RNAFOLD_PATH, FILE_ID) 
bulges_internal1, multiloops1, structures_recursive1 = main_function.compute_secondary_elements(stem_dic1, seq1, RNAFOLD_PATH, FILE_ID)
stems_shortened_dic1, pk_alignment_candidates1 = main_function.dotknot(seq1, stem_dic1, bulges_internal1, multiloops1, RNAFOLD_PATH, FILE_ID)
#---------------------------------------------------------------------------------------------------------
list_of_predictions = []
how_many_sequences = len(list_of_ids)

if how_many_sequences:
    for second_index in xrange(2, how_many_sequences + 1):
        j = second_index - 1
        pseudoknot_list, result, predicted_global_structure, score_sum, predicted_global_structure_seq2 = main_function.pairwise_prediction(stem_dic1, bulges_internal1, multiloops1, structures_recursive1, stems_shortened_dic1, pk_alignment_candidates1, 0, j, list_of_dictionaries, RNAFOLD_PATH, FILE_ID)
        
        print "Predicted structure for reference sequence folded with sequence ", list_of_ids[second_index - 1] 
        print seq1
        print predicted_global_structure
        print list_of_dictionaries[j]
        print predicted_global_structure_seq2 
        print "Score ", score_sum
        print
                          
        list_of_predictions.append((list_of_ids[second_index - 1], seq1, predicted_global_structure, list_of_dictionaries[j], predicted_global_structure_seq2, score_sum))
#---------------------------------------------------------------------------------------------------------
print '-------------------'
print
print "Predicted structures:"
print
print '-------------------' 
    
# Print structures with high scores first in file
if len(list_of_predictions) > 1:
    list_of_predictions = sorted(list_of_predictions, key=itemgetter(5), reverse = True)
 
for index, (identifier, seq1, structure, seq2, predicted_global_structure_seq2, score_sum) in enumerate(list_of_predictions):
    structure_list = list(structure)
    output_alignment.write(str(identifier))
    output_alignment.write('\n')    
    output_alignment.write(str(seq1))
    output_alignment.write('\n')          
    output_alignment.write(str(structure))
    output_alignment.write('\n')          
    output_alignment.write(str(seq2))
    output_alignment.write('\n')
    output_alignment.write(str(predicted_global_structure_seq2))
    output_alignment.write('\n')          
    output_alignment.write(str(score_sum))
    output_alignment.write('\n')    
output_alignment.close()

print 
print '-------------------'
print "Results were written to file ", path + "_output.txt"
print '-------------------'
print
#---------------------------------------------------------------------------------------------------------
try:
    os.remove(FILE_ID + '_stem_structure.txt')
    os.remove(FILE_ID + '_stacking_energy.txt')
    os.remove(FILE_ID + '_loops_energy.txt')
    os.remove(FILE_ID + '_input.fasta')
    os.remove(FILE_ID + '_bulge_internal_structures.txt')
    os.remove(FILE_ID + '_bulge_internal_energy.txt')
    os.remove(FILE_ID + '_multiloops_energy.txt')
    os.remove(FILE_ID + '_multiloop_structures.txt')
    os.remove(FILE_ID + '_ss.ps')
    os.remove(FILE_ID + '_dp.ps')
    os.remove(FILE_ID + '_dp2.ps')
except:
    pass
#---------------------------------------------------------------------------------------------------------        


