"""
    DotKnot-PW: software for pairwise RNA pseudoknot prediction
    Copyright (C) 2011-2012 Jana Sperschneider	

    This program is free software; you can redistribute it and/or modify  
    it under the terms of the GNU General Public License as published by  
    the Free Software Foundation; either version 2 of the License, or     
    (at your option) any later version.    
"""

import scores
from operator import itemgetter
#---------------------------------------------------------------------------------------------------------             
gev_parameters_regular_stems = {
# stemlength : location_gev, scale_gev, shape_gev               
3 : (6.2656300, 2.1964601, -0.1468699) ,
4 : (7.2147280, 2.4309990, -0.1406325) ,
5 : (8.116595, 2.514450, -0.116540) ,
6 : (7.99878347, 2.51635600, -0.08248844) ,
7 : (8.58702784, 2.40085386, -0.08390027) , 
8 : (9.97894899, 2.77863232, -0.07065527) ,
9 : (10.3113633, 2.54373517, -0.02892308) ,           
10: (10.3113633, 2.54373517, -0.02892308)
}
""" These are the extreme value parameters used for matching interrupted
    stem in first sequence matched with interrupted stem in second
    sequence """
gev_parameters_interrupted_stems = {
# stemlength : location_gev, scale_gev, shape_gev               
6 : (15.4001651, 3.1528310, -0.2444175) ,
7 : (17.4863558, 3.8084893, -0.2075487) ,
8 : (19.4195018, 4.3611266, -0.2251111) ,
9 : (21.0377466, 4.2479623, -0.1702904) ,           
10: (20.3211533, 4.9971441, -0.1947792) , 
11: (21.6764893, 5.6984451, -0.1800163) ,
12: (23.5446867, 5.7563755, -0.1836097) ,
13: (23.5523016, 6.0606109, -0.1197156) ,
14: (23.5523016, 6.0606109, -0.1197156) ,
15: (24.3302650, 6.6691585, -0.0078104) ,
16: (25.3617468, 6.8580243, -0.0465578) ,
17: (26.6491333, 7.5324883, -0.0058719) ,
18: (26.6491333, 7.5324883, -0.0058719) ,
19: (25.9293358, 7.3719887, 0.09594748) ,
20: (26.5253033, 7.3044115, 0.07121345) ,
21: (26.754421, 7.379724, 0.042155) ,
22: (25.980602, 7.126464, 0.106798) 
}
#---------------------------------------------------------------------------------------------------------     
def calculate_score_list_IB_r_IB(stem1, stemlength1, stem2_list, stemlengths2, ribosum_bases_first_seq, ribosum_bases_second_seq, ribosum_cache):    
    """ Function: calculate_score()

        Purpose:  Calculate RIBOSUM score.   
                  
        Input:    Two stems with values and precalculated RIBOSUM bases dictionaries. 
        
        Return:   RIBOSUM score.
    """    
    bps1 = ribosum_bases_first_seq[stem1, stemlength1]  
    
    score_list = []

    key1 = reduce(lambda x,y:10*x+y, bps1)
    
    for index, (stem2, values2) in enumerate(stem2_list):
        
        stemlength2 = stemlengths2[index]
        bps2 = ribosum_bases_second_seq[stem2, stemlength2]    
        
        key2 = reduce(lambda x,y:10*x+y, bps2)

        if (key1,key2) in ribosum_cache:
            ribosum_score = ribosum_cache[(key1,key2)]
        
        else:
            ribosum_score = scores.ribosum_score_DP(bps1, bps2)        
            ribosum_cache[(key1,key2)] = ribosum_score
                    
        score_list.append(ribosum_score)
     
    return score_list, ribosum_cache
#---------------------------------------------------------------------------------------------------------     
def calculate_score_list_IB_r(stem1, stemlength1, stem2_list, ribosum_bases_first_seq, ribosum_bases_second_seq, ribosum_cache):    
    """ Function: calculate_score()

        Purpose:  Calculate RIBOSUM score.   

                  
        Input:    Two stems with values and precalculated RIBOSUM bases dictionaries. 
        
        Return:   RIBOSUM score.
    """    
    bps1 = ribosum_bases_first_seq[stem1, stemlength1]  
    
    score_list = []

    key1 = reduce(lambda x,y:10*x+y, bps1)
    
    for (stem2, values2) in stem2_list:
        
        stemlength2 = values2[0]
        bps2 = ribosum_bases_second_seq[stem2, stemlength2]    
        
        key2 = reduce(lambda x,y:10*x+y, bps2)
        
        if (key1,key2) in ribosum_cache:
            ribosum_score = ribosum_cache[(key1,key2)]
        
        else:
            ribosum_score = scores.ribosum_score_DP(bps1, bps2)        
            ribosum_cache[(key1,key2)] = ribosum_score
                    
        score_list.append(ribosum_score)
     
    return score_list, ribosum_cache
#---------------------------------------------------------------------------------------------------------     

def conserved_bulges(first_seq, first_dic_stems, first_dic_bulges, second_seq, second_dic_stems, second_dic_bulges, P_VALUE, conserved_elements):
    """ Function: conserved_bulges()

        Purpose:  Two sequences with corresponding interrupted stem dictionaries
                  are given. Find significant pairs.                             
                  
        Input:    Two interrupted stem dictionaries and two sequences. 
        
        Return:   A dictionary of significant pairs.
    """     
    ribosum_cache = {}
    
    # Ribosum score pre-processing. Store list of base pairs for each stem
    ribosum_bases_first_seq = scores.ribosum_bases(first_seq, first_dic_stems, None, first_dic_bulges)
    ribosum_bases_second_seq = scores.ribosum_bases(second_seq, second_dic_stems, None, second_dic_bulges)    
 
    # Pre-calculate the RIBOSUM score thresholds with respect to given p-value    
    dic_gev_thresholds_regular = scores.thresholds(P_VALUE, gev_parameters_regular_stems)
    dic_gev_thresholds_interrupted = scores.thresholds(P_VALUE, gev_parameters_interrupted_stems)

    stemlengths1 = [values1[1].count('(') for stem1, values1 in first_dic_bulges.items()]  
    stemlengths2 = [values2[1].count('(') for stem2, values2 in second_dic_bulges.items()]  

    looplengths1 = [values1[1].count('.') for stem1, values1 in first_dic_bulges.items()]  
    looplengths2 = [values2[1].count('.') for stem2, values2 in second_dic_bulges.items()]     
     
    energies2 = [values2[2] for stem2, values2 in second_dic_bulges.items()]
    
    for index, (stem1, values1) in enumerate(first_dic_bulges.items()):

        stemlength1 = stemlengths1[index]
        looplength1 = looplengths1[index]
        energy1 = values1[2]     

        ribosum_score_list_IB_IB, ribosum_cache = calculate_score_list_IB_r_IB(stem1, stemlength1, second_dic_bulges.items(), stemlengths2, ribosum_bases_first_seq, ribosum_bases_second_seq, ribosum_cache)        
        prob_list_IB_IB = [scores.get_prob(stemlength1, ribosum_score_list_IB_IB[index], dic_gev_thresholds_interrupted, gev_parameters_interrupted_stems) for index, (stem2, values2) in enumerate(second_dic_bulges.items())]

        ribosum_score_list_IB_r, ribosum_cache = calculate_score_list_IB_r(stem1, stemlength1, second_dic_stems.items(), ribosum_bases_first_seq, ribosum_bases_second_seq, ribosum_cache)        
        prob_list_IB_r = [scores.get_prob(stemlength1, ribosum_score_list_IB_r[index], dic_gev_thresholds_regular, gev_parameters_regular_stems) for index, (stem2, values2) in enumerate(second_dic_stems.items())]
        
        """ Calculate score for interrupted stems and interrupted stems """

        for index, prob in enumerate(prob_list_IB_IB):
            if prob < P_VALUE:                
                stem2 = second_dic_bulges.items()[index][0]
                stemlength2 = stemlengths2[index] 
                energy2 = energies2[index]
                looplength2 = looplengths2[index]    
                ribosum_score = ribosum_score_list_IB_IB[index]     	                                              
                dissim = scores.dissimilarity(stemlength1, stemlength2, looplength1, looplength2)                                
                combined_score = scores.DP_score(ribosum_score, energy1 + energy2, dissim)                  
                if combined_score > 0.0:
                    conserved_elements[(stem1[0], stem1[1], 'ib', stem2[0], stem2[1], 'ib')] = combined_score, ribosum_score                                    
                    
        """ Calculate score for interrupted stems and regular stems """

        for index, prob in enumerate(prob_list_IB_r):        
            if prob < P_VALUE:
                stem2 = second_dic_stems.items()[index][0]          
                stemlength2 = second_dic_stems.items()[index][1][0]
                if scores.difference(stemlength1, stemlength2) <= 5:                                                
                    energy2 = second_dic_stems.items()[index][1][3]                
                    looplength2 = (stem2[1] - stemlength2) - (stem2[0] + stemlength2) + 1                    
                    ribosum_score = ribosum_score_list_IB_r[index]     	                                                  
                    dissim = scores.dissimilarity(stemlength1, stemlength2, looplength1, looplength2)                                
                    combined_score = scores.DP_score(ribosum_score, energy1 + energy2, dissim)                  
                    if combined_score > 0.0:
                        conserved_elements[(stem1[0], stem1[1], 'ib', stem2[0], stem2[1], 'hp')] = combined_score, ribosum_score                            
                              
    for stem1, values1 in first_dic_stems.items():
        
        stemlength1 = values1[0] 
        energy1 = values1[3]   
        looplength1 = (stem1[1] - stemlength1) - (stem1[0] + stemlength1) + 1           
        ribosum_score_list_r_IB, ribosum_cache = calculate_score_list_IB_r_IB(stem1, stemlength1, second_dic_bulges.items(), stemlengths2, ribosum_bases_first_seq, ribosum_bases_second_seq, ribosum_cache)        
        prob_list_r_IB = [scores.get_prob(stemlength1, ribosum_score_list_r_IB[index], dic_gev_thresholds_regular, gev_parameters_regular_stems) for index, (stem2, values2) in enumerate(second_dic_bulges.items())]                
        
        """ Calculate score for regular stems and interrupted stems """

        for index, prob in enumerate(prob_list_r_IB):                        
            if prob < P_VALUE:                       
                stem2 = second_dic_bulges.items()[index][0]         
                stemlength2 = stemlengths2[index] 
                if scores.difference(stemlength1, stemlength2) <= 5:  
                    energy2 = energies2[index]
                    looplength2 = looplengths2[index]    
                    ribosum_score = ribosum_score_list_r_IB[index]     	                                                  
                    dissim = scores.dissimilarity(stemlength1, stemlength2, looplength1, looplength2)                                
                    combined_score = scores.DP_score(ribosum_score, energy1 + energy2, dissim)                  
                    if combined_score > 0.0:
                        conserved_elements[(stem1[0], stem1[1], 'hp', stem2[0], stem2[1], 'ib')] = combined_score, ribosum_score                        
                    
      
    return conserved_elements
    
  
def conserved_multiloops(first_seq, first_dic_mls, second_seq, second_dic_mls, structures_recursive1, structures_recursive2, conserved_stems, conserved_elements, P_VALUE):
    """ Function: conserved_multiloops()

        Purpose:  Two sequences with corresponding multiloop dictionaries
                  are given. For each multiloop in first sequence, find
                  significant multiloop pairs in second sequence.                             
                  
        Input:    Two multiloop dictionaries. 
        
        Return:   A dictionary of significant multiloop pairs.
    """          
    # Sort list of multiloops by right endpoint
    multiloops_seq1 = sorted(first_dic_mls, key=itemgetter(1,0))
    multiloops_seq2 = sorted(second_dic_mls, key=itemgetter(1,0))

    dic_gev_thresholds_interrupted = scores.thresholds(P_VALUE, gev_parameters_interrupted_stems)
          
    for ml1 in multiloops_seq1:

        probs_gev = []
        
        stemlength1 = first_dic_mls[ml1][0]
        length1 = ml1[1] - ml1[0] + 1
                
        inner_stems1 = structures_recursive1[(ml1[0], ml1[1], stemlength1)]
        # Now check if inner stems are interrupted stems or multiloops and if so, 
        # combine list of nested stems into one interrupted stem key        
        inner_stems1_structures = inner_stem_convert(inner_stems1)
        
        for ml2 in multiloops_seq2:                        
            stemlength2 = second_dic_mls[ml2][0]
            length2 = ml2[1] - ml2[0] + 1
            
            inner_stems2 = structures_recursive2[(ml2[0], ml2[1], stemlength2)]
            # Now check if inner stems are interrupted stems or multiloops and if so, 
            # combine list of nested stems into one interrupted stem key            
            inner_stems2_structures = inner_stem_convert(inner_stems2)        
            
            outer_score, score, combined_score = 0.0, 0.0, 0.0
            
            # Pure similarity conservation score for outer stem
            if (ml1, ml2) in conserved_stems:                
                outer_score = conserved_stems[ml1, ml2][1]                            
            
            if outer_score:                                    
                
                # Now calculate inner score match with local alignment strategy
                inner_score_sum = scores.inner_score_DP(inner_stems1_structures, inner_stems2_structures, conserved_elements)                                                
                if inner_score_sum: 

                    ribosum_score = outer_score + inner_score_sum

                    prob = scores.get_prob(stemlength1, ribosum_score, dic_gev_thresholds_interrupted, gev_parameters_interrupted_stems)
                    
                    if prob < P_VALUE:

                        stemlength1 = first_dic_mls[ml1][1].count('(')
                        stemlength2 = second_dic_mls[ml2][1].count('(')
                        looplength1 = first_dic_mls[ml1][1].count('.')
                        looplength2 = second_dic_mls[ml2][1].count('.')
                        energy1 = first_dic_mls[ml1][2]
                        energy2 = second_dic_mls[ml2][2]  
                                                    
                        dissim = scores.dissimilarity(stemlength1, stemlength2, looplength1, looplength2)
                        
                        combined_score = scores.DP_score(ribosum_score, energy1 + energy2, dissim)

                        if combined_score > 0.0:
                            conserved_elements[(ml1[0], ml1[1], 'ml', ml2[0], ml2[1], 'ml')] = combined_score, prob
                
    return conserved_elements


def inner_stem_convert(list_of_stems):
    """ Function: inner_stem_convert()

        Purpose: Given a list of nested structure elements for an outer stem
                 of a multiloop structure, convert it into a list of nested
                 secondary structure elements. For example, convert
                 [(100, 138, 4, 0.85900841402483996, 2.6837555719550492),
                  (105, 133, 5, 0.92199083901095125, 1.8247471579302095)]
                 to (100, 138, 'ib') as it represents an interrupted stem.
        Input:    
        
        Return:   
    """ 
    inner_stems_structures = []
    inner_stems_as_keys = []
    
    for position, stem in enumerate(list_of_stems):

        marker_stem = False
        marker_interrupted = False
        marker_multiloop = False        
        
        # Check if current stem is a plain stem
        marker_stem = is_it_a_stem(stem, list_of_stems[position + 1:])
        if marker_stem:
            tuple = stem[0], stem[1], stem[2], stem[3], stem[4], 'hp'
            inner_stems_as_keys.append(tuple)

        else:
            # Check if current stem is an interrupted stem
            marker_interrupted = is_it_an_interrupted_stem(stem, list_of_stems[position + 1:])
            if marker_interrupted:
                tuple = stem[0], stem[1], stem[2], stem[3], stem[4], 'ib'
                inner_stems_as_keys.append(tuple)
            else:
                tuple = stem[0], stem[1], stem[2], stem[3], stem[4], 'ml'
                inner_stems_as_keys.append(tuple)

    inner_stems = inner_stems_as_keys[:]
    
    # Now combine everything into keys for secondary structure elements
    for position, stem in enumerate(inner_stems_as_keys):
        if stem[-1] == 'hp':
            pass

        elif stem[-1] == 'ib':
            for inner_stem in inner_stems_as_keys[position + 1:]:
                if stem[0] < inner_stem[0] and inner_stem[1] < stem[1]:
                    if inner_stem in inner_stems:                                              
                        inner_stems.remove(inner_stem)                      
            
        elif stem[-1] == 'ml':
            for inner_stem in inner_stems_as_keys[position + 1:]:
                if stem[0] < inner_stem[0] and inner_stem[1] < stem[1]:
                    if inner_stem in inner_stems:                                              
                        inner_stems.remove(inner_stem)                    
    
    return inner_stems
    

def is_it_a_stem(stem, list_of_stems):

    marker_stem = False
    
    if list_of_stems:
        next_stem = list_of_stems[0]
        if stem[1] < next_stem[0]:
            marker_stem = True
    else:
        marker_stem = True
        
    return marker_stem
    

def is_it_an_interrupted_stem(stem, list_of_stems):

    marker_interrupted = False      

    left_boundary = stem[0]
    right_boundary = stem[1]
    
    for stem_inner in list_of_stems:
        if stem_inner[0] < stem[1]:
            if stem_inner[0] > left_boundary and stem_inner[1] < right_boundary:
                marker_interrupted = True
                left_boundary = stem_inner[0]
                right_boundary = stem_inner[1]
            else:
                marker_interrupted = False                    

    return marker_interrupted
    
